@@ -19,6 +19,7 @@ import pygments
 from pygments import formatter, lexers
 import zipwrap
 import Image
+import imagescale
 
 DOC_CONTENT_ATTRIB = {
     'office:version': '1.0',
@@ -58,6 +59,10 @@ for key, value in DOC_CONTENT_ATTRIB.items():
 TEXT_COUNT = 100
 DATA_DIR =  os.path.join(os.path.dirname(__file__), 'data')
 
+MONO_FONT = 'Courier New' # I like 'Envy Code R'
+
+SLIDE_WIDTH = 30 # cm
+SLIDE_HEIGHT = 21 
 
 def cwd_decorator(func):
     """
@@ -113,10 +118,18 @@ def to_xml(node):
     xml = fout.getvalue()
     return xml
 
-def pretty_xml(string_input):
+def pretty_xml(string_input, add_ns=False):
     """ pretty indent string_input """
+    if add_ns:
+        elem = '<foo '
+        for key, value in DOC_CONTENT_ATTRIB.items():
+            elem += ' %s="%s"' %(key, value)
+        string_input = elem + '>' + string_input + '</foo>'
     doc = minidom.parseString(string_input)
-    s1 = doc.toprettyxml('  ')
+    if add_ns:
+        s1 = doc.childNodes[0].childNodes[0].toprettyxml('  ')
+    else:
+        s1 = doc.toprettyxml('  ')
     return s1
 
 
@@ -125,6 +138,7 @@ class Preso(object):
 
     def __init__(self):
         self.slides = []
+        self.limit_pages = [] # can be list of page numbers (not indexes to export)
         self._pictures = [] # list of Picture instances
         self._footer_count = 0
         # xml elements 
@@ -142,6 +156,23 @@ class Preso(object):
         self._auto_styles = sub_el(self._root, 'office:automatic-styles')
         o_body = sub_el(self._root, 'office:body')
         self._presentation = sub_el(o_body, 'office:presentation')
+        
+    def add_imported_auto_style(self, style_node):
+        self._auto_styles.append(style_node)
+        style_node.parent = self._auto_styles
+
+    def import_slide(self, preso_file, page_num):
+        odp = zipwrap.ZipWrap(preso_file)
+        content = odp.cat('content.xml')
+        content_tree = et.fromstring(content)
+        slides = content_tree.findall('{urn:oasis:names:tc:opendocument:xmlns:office:1.0}body/{urn:oasis:names:tc:opendocument:xmlns:office:1.0}presentation/{urn:oasis:names:tc:opendocument:xmlns:drawing:1.0}page')
+        try:
+            slide_xml = slides[page_num - 1]
+        except IndexError, e:
+            print "Can't find page_num %d only %d slides" %(page_num, len(slides))
+            raise
+        if slide_xml:
+            self.slides.append(XMLSlide(self, slide_xml, odp))
 
     def get_data(self, style_file=None):
         fd, filename = tempfile.mkstemp()
@@ -191,12 +222,22 @@ class Preso(object):
 <manifest:manifest xmlns:manifest="urn:oasis:names:tc:opendocument:xmlns:manifest:1.0">
  <manifest:file-entry manifest:media-type="application/vnd.oasis.opendocument.presentation" manifest:full-path="/"/>
 """
-        for filename in zip.cat('/'):
+        files = zip.cat('/')
+        try:
+            files.extend(zip.cat('/Pictures'))
+        except IOError, e:
+            # it's ok to not have pictures ;)
+            pass
+        for filename in files:
             filetype = ''
             if filename.endswith('.xml'):
                 filetype = 'text/xml'
             elif filename.endswith('.jpg'):
                 filetype = 'image/jpeg'
+            elif filename.endswith('.gif'):
+                filetype = 'image/gif'
+            elif filename.endswith('.png'):
+                filetype = 'image/png'
             elif filename == 'Configurations2/':
                 filetype = 'application/vnd.sun.xml.ui.configuration'
             
@@ -234,6 +275,9 @@ class Preso(object):
 
     def to_xml(self):
         for i, slide in enumerate(self.slides):
+            if self.limit_pages and i+1 not in self.limit_pages:
+
+                continue
             if slide.footer:
                 footer_node = slide.footer.get_node()
                 self._presentation.append(footer_node)
@@ -304,6 +348,19 @@ class Animation(object):
                                                     'smil:to':'visible'})
         return par
 
+                 
+class ImportedPicture(object):
+    """
+    Pictures used when importing slides
+    """
+    def __init__(self, name, data):
+        self.internal_name = name
+        self.data = data
+        
+    def get_data(self):
+        return self.data
+
+
 class Picture(object):
     """
     Need to convert to use image scale::
@@ -325,11 +382,27 @@ class Picture(object):
     COUNT = 0
     CM_SCALE = 30.
 
-    def __init__(self, filepath):
+    def __init__(self, filepath, **kw):
         self.filepath = filepath
         image = Image.open(filepath)
         self.w, self.h = image.size
         self.internal_name = self._gen_name()
+        self.user_defined = {}
+        self._process_kw(kw)
+
+    def update_frame_attributes(self, attrib):
+        """ For positioning update the frame """
+            
+        if 'align' in self.user_defined:
+            align = self.user_defined['align']
+            if 'top' in align:
+                attrib['style:vertical-pos'] = 'top'
+            if 'right' in align:
+                attrib['style:horizontal-pos'] = 'right'
+        return attrib
+
+    def _process_kw(self, kw):
+        self.user_defined = kw
 
     def _gen_name(self):
         ext = os.path.splitext(self.filepath)[1]
@@ -337,21 +410,116 @@ class Picture(object):
         Picture.COUNT += 1
         return name
     
-    def get_width(self, measurement):
-        if measurement == 'cm':
+    def get_xywh(self, measurement=None):
+        if measurement is None or measurement == 'cm':
+            measurement = 'cm'
             scale = Picture.CM_SCALE
-        return str(self.w/scale)+measurement
+        
+        DPCM = 1 # dots per cm
+        if 'crop' in self.user_defined.get('classes', []):
+            x,y,w,h = imagescale.adjust_crop(SLIDE_WIDTH*DPCM, SLIDE_HEIGHT*DPCM,self.w, self.h)
+        elif 'fit' in self.user_defined.get('classes', []):
+            x,y,w,h = imagescale.adjust_fit(SLIDE_WIDTH*DPCM, SLIDE_HEIGHT*DPCM,self.w, self.h)
+        elif 'fill' in self.user_defined.get('classes', []):
+            x,y,w,h = 0,0,SLIDE_WIDTH,SLIDE_HEIGHT
+        else:
+            x,y,w,h = 1.4, 4.6, self.get_width(), self.get_height()
+        return [str(foo)+measurement for foo in [x,y,w,h]]
 
-    def get_height(self, measurement):
+    def get_width(self, measurement=None):
+        if measurement is None or measurement == 'cm':
+            measurement = 'cm'
+            scale = Picture.CM_SCALE
+        if 'width' in self.user_defined:
+            return self.user_defined['width'] + 'pt'
+        if 'scale' in self.user_defined:
+            return '%spt' % (self.w * float(self.user_defined['scale'])/100)
+        return str(self.w/scale)
+
+    def get_height(self, measurement=None):
+        if measurement is None:
+            measurement = 'cm'
         if measurement == 'cm':
             scale = Picture.CM_SCALE
-        return str(self.h/scale)+measurement
+        if 'height' in self.user_defined:
+            return self.user_defined['height'] + 'pt'
+        if 'scale' in self.user_defined:
+            return '%spt' % (self.h * float(self.user_defined['scale'])/100)
+        return str(self.h/scale)
 
 
 
     def get_data(self):
         return open(self.filepath).read()
 
+class XMLSlide(object):
+    PREFIX = 'IMPORT%d-%s'
+    COUNT = 0
+    def __init__(self, preso, node, odp_zipwrap):
+        self.preso = preso
+        self.page_node = node
+        self.footer = None
+        self.mangled = self._mangle_name()
+        self._init(odp_zipwrap)
+
+    def page_num(self):
+        """ not an int, usually 'Slide 1' or 'page1' """
+        name = self.page_node.attrib.get('{urn:oasis:names:tc:opendocument:xmlns:drawing:1.0}name', None) 
+        return name
+
+    def _mangle_name(self):
+        name = self.PREFIX%(self.COUNT, self.page_num())
+        self.COUNT += 1
+        return name
+
+    def _init(self, odp_zipwrap):
+
+        # pull pictures out of slide
+        images = self.page_node.findall('*/{urn:oasis:names:tc:opendocument:xmlns:drawing:1.0}image')
+        for image in images:
+            path = image.attrib.get('{http://www.w3.org/1999/xlink}href')
+            data = odp_zipwrap.cat(path)
+            name = path.split('/')[1]
+            self.preso._pictures.append(ImportedPicture(name, data))
+
+        # pull styles out of content.xml (draw:style-name, draw:text-style-name, text:style-name)
+        styles_to_copy = {} #map of (attr_name, value) to value
+        attr_names = ['{urn:oasis:names:tc:opendocument:xmlns:drawing:1.0}style-name',
+                      '{urn:oasis:names:tc:opendocument:xmlns:drawing:1.0}text-style-name',
+                      '{urn:oasis:names:tc:opendocument:xmlns:text:1.0}style-name']
+        for node in self.page_node.getiterator():
+            for attr_name in attr_names:
+                style = node.attrib.get(attr_name, None)
+                if style:
+                    styles_to_copy[style] = attr_name
+                    # mangle name
+                    node.attrib[attr_name] = self.mangled + style
+
+        auto_attr_names = ['{urn:oasis:names:tc:opendocument:xmlns:style:1.0}name']
+        found = {}
+        # get content.xml automatic-styles
+        content = odp_zipwrap.cat('content.xml')
+        content_node = et.fromstring(content)
+        auto_node = content_node.findall('{urn:oasis:names:tc:opendocument:xmlns:office:1.0}automatic-styles')[0]
+
+        for node in auto_node.getchildren():
+            for attr_name in auto_attr_names:
+                attr_value = node.attrib.get(attr_name, None)
+                if  attr_value in styles_to_copy:
+                    found[attr_value] = 1
+                    # mangle name
+                    node.attrib[attr_name] = self.mangled + attr_value
+                    self.preso.add_imported_auto_style(node)
+        
+        
+        
+        
+
+    def get_node(self):
+        return self.page_node                               
+
+        
+                          
 class Slide(object):
     def __init__(self, preso, page_number=None):
         self.title_frame = None
@@ -373,11 +541,21 @@ class Slide(object):
                                 # text or notes (Subclass of
                                 # MixedContent)
 
+        self.insert_line_break = 0
+
         # xml elements
         self._page = None
 
         self._init_xml()
 
+    def insert_line_breaks(self):
+        """
+        If you want to write out existing line breaks, but don't have content to write 
+        call this
+        """
+        if self.cur_element:
+            self.cur_element.line_break()
+
     def start_animation(self, anim):
         self.animations.append(anim)
         self.paragraph_attribs['text:id'] = anim.id
@@ -385,7 +563,19 @@ class Slide(object):
     def end_animation(self):
         # jump out of text:p
         self.parent_of('text:p')
-        del self.paragraph_attribs['text:id']
+        if 'text:id' in self.paragraph_attribs:
+            del self.paragraph_attribs['text:id']
+
+    def push_pending_node(self, name, attr):
+        """
+        pending nodes are for affecting type, such as wrapping content
+        with text:a to make a hyperlink.  Anything in pending nodes
+        will be written before the actual text.
+        User needs to remember to pop out of it.
+        """
+        if self.cur_element is None:
+            self.add_text_frame()
+        self.cur_element.pending_nodes.append((name,attr))
 
     def push_style(self, style):
         if self.cur_element is None:
@@ -427,6 +617,8 @@ class Slide(object):
         node.parent = self._page
 
     def push_element(self):
+        """ element push/pop is used to remember previous cur_elem, since
+        lists might need to mess with that"""
         self.element_stack.append(self.cur_element)
 
     def pop_element(self):
@@ -528,13 +720,12 @@ class Slide(object):
             self.add_text_frame()
         self.cur_element.write(text, **kw)
 
-    def add_node(self, node, attrib):
+    def add_node(self, node, attrib=None):
+        attrib = attrib or {}
         if self.cur_element is None:
             self.add_text_frame()
         self.cur_element.add_node(node, attrib)
 
-    def add_p(self):
-        self.cur_element.add_p()
 
     def pop_node(self):
         self.cur_element.pop_node()
@@ -549,18 +740,22 @@ class Slide(object):
 
 class MixedContent(object):
     def __init__(self, slide, name, attrib=None):
+        self._default_align = 'start'
         self.slide = slide
         if attrib is None:
             attrib = {}
         self.node = el(name, attrib)
         self.cur_node = self.node
         self.pending_styles = []
+        # store nodes that affect output (such as text:a)
+        self.pending_nodes = [] # typles of (name, attr)
+        self.dirty = False # keep track if we have been written to
 
     def parent_of(self, name):
         """ 
         go to parent of node with name, and set as cur_node.  Useful
         for creating new paragraphs
-        """
+       """
         if not self._in_tag(name):
             return
         node = self.cur_node
@@ -575,16 +770,24 @@ class MixedContent(object):
         """
         return self._in_tag('text:p')
 
+    def _is_last_child(self, tagname, attributes=None):
+        """
+        Check if last child of cur_node is tagname with attributes
+        """
+        children = self.cur_node.getchildren()
+        if children:
+            result = self._is_node(tagname, attributes, node=children[-1])
+            return result
+        return False
+        
     def _is_node(self, tagname, attributes=None, node=None):
-        node = node or self.cur_node
+        if node is None:
+            node = self.cur_node
         if attributes:
             return node.tag == tagname and node.attrib == attributes
         else:
             return node.tag == tagname
 
-    def _is_parent(self, tagname, attributes=None):
-        return self._is_node(tagname, attributes, self.cur_node.parent)
-
     def _in_tag(self, tagname, attributes=None):
         """
         Determine if we are already in a certain tag.
@@ -611,7 +814,17 @@ class MixedContent(object):
         self.cur_node.append(node)
         node.parent = self.cur_node
     
+    def _check_add_node(self, parent, name):
+        ''' Returns False if bad to make name a child of parent '''
+
+        if name == 'text:a':
+            if parent.tag == 'draw:text-box':
+                return False
+        return True
+        
     def _add_node(self, parent, name, attrib):
+        if not self._check_add_node(parent, name):
+            raise Exception, 'Bad child (%s) for %s)' %(name, parent.tag)
         new_node = sub_el(parent, name, attrib)
         return new_node
         
@@ -643,7 +856,10 @@ class MixedContent(object):
         if add_paragraph or self.slide.paragraph_attribs:
             p_attrib = {'text:style-name':para.name}
             p_attrib.update(self.slide.paragraph_attribs)
-            if not self._in_tag('text:p', p_attrib):
+            if self._is_last_child('text:p', p_attrib):
+                children = self.cur_node.getchildren()
+                self.cur_node = children[-1] 
+            elif not self._in_tag('text:p', p_attrib):
                 self.parent_of('text:p')
                 # Create paragraph style first
                 self.slide._preso.add_style(para)
@@ -653,11 +869,40 @@ class MixedContent(object):
         if add_text:
             if t_styles:
                 text = TextStyle(**t_styles)
+                children = self.cur_node.getchildren()
+                if children:
+                    # if we already are using this text style, reuse the last one
+                    last = children[-1]
+                    if last.tag == 'text:span' and last.attrib['text:style-name'] == text.name:
+                        self.cur_node = children[-1]
+                        return 
                 if not self._is_node('text:span', {'text:style-name':text.name}):
                     # Create text style
                     self.slide._preso.add_style(text)
                     self.add_node('text:span', attrib={'text:style-name':text.name})
 
+
+    def _add_pending_nodes(self):
+        for node, attr in self.pending_nodes:
+            self.add_node(node, attr)
+            
+
+    def line_break(self):
+        """insert as many line breaks as the insert_line_break variable says
+        """
+        for i in range(self.slide.insert_line_break):
+            # needs to be inside text:p
+            if not self._in_tag('text:p'):
+                # we can just add a text:p and no line-break
+                # Create paragraph style first
+                self.add_node('text:p')
+            else:
+                self.add_node('text:line-break')
+            self.pop_node()
+
+        self.slide.insert_line_break = 0
+
+
     def write(self, text, add_p_style=True, add_t_style=True):
         """
         see mixed content
@@ -667,41 +912,60 @@ class MixedContent(object):
         white spaces then dealing with the '' (empty strings) which
         would be the extra spaces
         """
-
+        self.line_break()
         self._add_styles(add_p_style, add_t_style)
+        self._add_pending_nodes()
 
-        seen_space = False
-        for chunk in text.split(' '): # deal with white space 
-            # see  http://books.evc-cit.info/odbook/ch03.html#whitespace-section 
-            if chunk == '' and not seen_space:
-                seen_space = True
-                #chunk = ' '
-                #continue
-                # skip the first space
-            elif chunk == '':
-                self.add_node('text:s', {})
-                self.pop_node()
+        spaces = []
+        for i, letter in enumerate(text): 
+            if letter == ' ':
+                spaces.append(letter)
                 continue
+            elif len(spaces) == 1:
+                self._write(' ')
+                self._write(letter)
+                spaces = []
+                continue
+            elif spaces:
+                num_spaces = len(spaces) - 1
+                # write just a plain space at the start
+                self._write(' ')
+                if num_spaces > 1:
+                    # write the attrib only if more than one space
+                    self.add_node('text:s', {'text:c':str(num_spaces)})
+                else:
+                    self.add_node('text:s')
+                self.pop_node()
+                self._write(letter)
+                spaces = []
+                continue
+            self._write(letter)
+
+        # might have dangling spaces
+        if len(spaces) == 1:
+            self._write(' ')
+        elif spaces:
+        #if spaces:
+            num_spaces = len(spaces) - 1
+            # write space
+            self._write(' ')
+            if num_spaces > 1:
+                self.add_node('text:s', {'text:c':str(num_spaces)})
             else:
-                seen_space = False
-
-            children = self.cur_node.getchildren()
-            #if self.cur_node.text and children:
-            if children:
-                child = children[-1]
-                cur_text = child.tail or ''
-                if chunk == '' or (cur_text and cur_text[-1] != ' '):
-                    # need to put ONE space in for initial spaces
-                    # or in between chunks
-                    chunk = ' ' + chunk
-                child.tail = cur_text + chunk
-            else:    
-                cur_text = self.cur_node.text or ''
-                if chunk == '' or (cur_text and cur_text[-1] != ' '):
-                    # need to put a space in for initial spaces
-                    chunk = ' ' + chunk
-                self.cur_node.text = cur_text + chunk
+                self.add_node('text:s')
+            self.pop_node()
+            
 
+    def _write(self, letter):
+        children = self.cur_node.getchildren()
+        if children:
+            child = children[-1]
+            cur_text = child.tail or ''
+            child.tail = cur_text + letter
+        else:    
+            cur_text = self.cur_node.text or ''
+            self.cur_node.text = cur_text + letter
+        self.dirty = True
 
 class Footer(MixedContent):
     def __init__(self, slide):
@@ -719,15 +983,17 @@ class Footer(MixedContent):
 
 class PictureFrame(MixedContent):
     def __init__(self, slide, picture, attrib=None):
+        x,y,w,h = picture.get_xywh()
         attrib = attrib or {
             'presentation:style-name':'pr2',
             'draw:style-name':'gr2',
             'draw:layer':'layout',
-            'svg:width':picture.get_width('cm'),
-            'svg:height':picture.get_height('cm'),
-            'svg:x':'1.4cm',
-            'svg:y':'4.577cm',
+            'svg:width':w, #picture.get_width(),
+            'svg:height':h, #picture.get_height(),
+            'svg:x':x, #'1.4cm',
+            'svg:y':y, #'4.577cm',
             }
+        attrib = picture.update_frame_attributes(attrib)
         MixedContent.__init__(self, slide, 'draw:frame', attrib=attrib)
         
 
@@ -749,7 +1015,7 @@ class TextFrame(MixedContent):
         self.text_styles = ['P1']
 
         self.cur_node = self._text_box
-        self._default_align = 'start'
+
 
     def to_xml(self):
         return to_xml(self.get_node())
@@ -757,18 +1023,6 @@ class TextFrame(MixedContent):
     def _in_bullet(self):
         return self._in_tag('text:list')
 
-    def add_p(self):
-        if not self._in_p():
-            style = ParagraphStyle(**{'fo:text-align':self._default_align,
-                                      'fo:margin-left':'1.2cm',
-                                      'fo:margin-right':'-.9cm',})
-            self.slide.push_style(style)
-
-
-    def generate_style(self):
-        ''' return the current text style name and add any automatic styles
-        needed (not generated yet)'''
-        return self.text_styles[-1]
         
 class TitleFrame(TextFrame):
     def __init__(self, slide, attrib=None):
@@ -949,17 +1203,25 @@ class OdtCodeFormatter(formatter.Formatter):
     def format(self, source, outfile):
         tclass = pygments.token.Token
         for ttype, value in source:
+            # getting ttype, values like (Token.Keyword.Namespace, u'')
+            if value == '':
+                continue
             style_attrib = self.get_style(ttype)
             tstyle = TextStyle(**style_attrib)
             self.writable.slide.push_style(tstyle)
-            parts = value.split('\n')
-            for part in parts[:-1]:
-                self.writable.write(part)
-                self.writable.add_node('text:line-break', {})
-                self.writable.pop_node()
-            self.writable.write(parts[-1])
+            if value == '\n':
+                self.writable.slide.insert_line_break = 1
+                self.writable.line_break()
+            else:
+                parts = value.split('\n')
+                for part in parts[:-1]:
+                    self.writable.write(part)
+                    self.writable.slide.insert_line_break = 1
+                    self.writable.line_break()
+                self.writable.write(parts[-1])
             self.writable.slide.pop_style()
             self.writable.pop_node()
+
             
     def get_style(self, tokentype):
         while not self.style.styles_token(tokentype):
@@ -967,7 +1229,7 @@ class OdtCodeFormatter(formatter.Formatter):
         value = self.style.style_for_token(tokentype)
         # default to monospace
         results = {
-            'fo:font-family':"Courier New",
+            'fo:font-family':MONO_FONT,
             'style:font-family-generic':"swiss",
             'style:font-pitch':"fixed"}
         if value['color']:
@@ -1031,6 +1293,9 @@ http://books.evc-cit.info/odbook/ch03.html#bulleted-numbered-lists-section
         self._default_align = 'start'
         self.attrib = attrib or {'text:style-name':'L2'}
         MixedContent.__init__(self, slide, 'text:list', attrib=self.attrib)
+        # if slide already has text insert line break
+        self.line_break()
+
         self.parents = [self.node]
         self.level = 0
         self.style_file = 'auto_list.xml'
